
#include "fw/dynmx/sequencer.h"
#include "it.h"
#include "main.h"
#include "setup.h"
#include "stm32g4xx_hal.h"

// alternative: use idle itnerrupt that is fired if Receive happend but is waiting for rest of the
// bytes \o/ hint:
// https://github.com/SteelPh0enix/STM32_HM10_Driver/blob/master/Drivers/HM-10/hm10.cpp#L55
// https://github.com/SteelPh0enix/STM32_HM10_Driver/blob/master/Core/Src/freertos.cpp#L247

using namespace schpin;

perif_handlers HANDLERS;

dynmx_sequencer< dynmx_master_message >             lm_seq;
std::array< uint8_t, decltype( lm_seq )::max_size > lm_buff;
dynmx_master_message                                lm_out;

dynmx_sequencer< master_dynmx_message >             ml_seq;
std::array< uint8_t, decltype( ml_seq )::max_size > ml_buff;
master_dynmx_message                                ml_out;

void start_cycle()
{
        auto [req_b] = ml_seq.start_read_packet();
        HAL_UART_Receive_IT( &HANDLERS.com_uart, ml_buff.data(), static_cast< uint16_t >( req_b ) );
}

void __gnu_cxx::__verbose_terminate_handler()
{
        std::abort();
}

int main()
{
        HANDLERS = setup_system();

        HAL_UART_RegisterCallback(
            &HANDLERS.bus_uart, HAL_UART_TX_COMPLETE_CB_ID, []( UART_HandleTypeDef* ) {
                    auto [req_b] = lm_seq.start_read_packet();
                    HAL_HalfDuplex_EnableReceiver( &HANDLERS.bus_uart );
                    HAL_UART_Receive_IT(
                        &HANDLERS.bus_uart, lm_buff.data(), static_cast< uint16_t >( req_b ) );
            } );
        HAL_UART_RegisterCallback(
            &HANDLERS.bus_uart, HAL_UART_RX_COMPLETE_CB_ID, []( UART_HandleTypeDef* u ) {
                    lm_seq.isr_read_callback( em::view_n( lm_buff.begin(), u->RxXferSize ) )
                        .match(
                            [&]( const dynmx_read_request& req ) {
                                    HAL_UART_Receive_IT(
                                        &HANDLERS.bus_uart,
                                        lm_buff.data(),
                                        static_cast< uint16_t >( req.count ) );
                            },
                            [&]( const dynmx_master_message& m ) {
                                    lm_out = m;
                                    HAL_UART_Transmit_IT(
                                        &HANDLERS.com_uart,
                                        lm_out.begin(),
                                        static_cast< uint16_t >( lm_out.size() ) );
                            } );
            } );

        HAL_UART_RegisterCallback(
            &HANDLERS.com_uart, HAL_UART_TX_COMPLETE_CB_ID, []( UART_HandleTypeDef* ) {
                    start_cycle();
            } );
        HAL_UART_RegisterCallback(
            &HANDLERS.com_uart, HAL_UART_RX_COMPLETE_CB_ID, []( UART_HandleTypeDef* u ) {
                    ml_seq.isr_read_callback( em::view_n( ml_buff.begin(), u->RxXferSize ) )
                        .match(
                            [&]( const dynmx_read_request& req ) {
                                    HAL_UART_Receive_IT(
                                        &HANDLERS.com_uart,
                                        ml_buff.data(),
                                        static_cast< uint16_t >( req.count ) );
                            },
                            [&]( const master_dynmx_message& m ) {
                                    ml_out = m;
                                    HAL_HalfDuplex_EnableTransmitter( &HANDLERS.bus_uart );
                                    HAL_UART_Transmit_IT(
                                        &HANDLERS.bus_uart,
                                        ml_out.begin(),
                                        static_cast< uint16_t >( ml_out.size() ) );
                            } );
            } );

        start_cycle();
        while ( true ) {
                asm( "nop" );
        }
}
